MODULE limupdate2
   !!======================================================================
   !!                     ***  MODULE  limupdate2  ***
   !!   LIM-3 : Update of sea-ice global variables at the end of the time step
   !!======================================================================
   !! History :  3.0  !  2006-04  (M. Vancoppenolle) Original code
   !!----------------------------------------------------------------------
#if defined key_lim3
   !!----------------------------------------------------------------------
   !!   'key_lim3'                                      LIM3 sea-ice model
   !!----------------------------------------------------------------------
   !!    lim_update2   : computes update of sea-ice global variables from trend terms
   !!----------------------------------------------------------------------
   USE limrhg          ! ice rheology

   USE dom_oce
   USE oce             ! dynamics and tracers variables
   USE in_out_manager
   USE sbc_oce         ! Surface boundary condition: ocean fields
   USE sbc_ice         ! Surface boundary condition: ice fields
   USE dom_ice
   USE phycst          ! physical constants
   USE ice
   USE limdyn
   USE limtrp
   USE limthd
   USE limsbc
   USE limdiahsb
   USE limwri
   USE limrst
   USE thd_ice         ! LIM thermodynamic sea-ice variables
   USE par_ice
   USE limitd_th
   USE limitd_me
   USE limvar
   USE prtctl           ! Print control
   USE lbclnk           ! lateral boundary condition - MPP exchanges
   USE wrk_nemo         ! work arrays
   USE lib_fortran     ! glob_sum
   USE timing          ! Timing

   IMPLICIT NONE
   PRIVATE

   PUBLIC   lim_update2   ! routine called by ice_step

      REAL(wp)  ::   epsi10 = 1.e-10_wp   !    -       -
      REAL(wp)  ::   rzero  = 0._wp       !    -       -
      REAL(wp)  ::   rone   = 1._wp       !    -       -
         
   !! * Substitutions
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !! NEMO/LIM3 4.0 , UCL - NEMO Consortium (2011)
   !! $Id: limupdate.F90 3294 2012-01-28 16:44:18Z rblod $
   !! Software governed by the CeCILL licence     (NEMOGCM/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------
CONTAINS

   SUBROUTINE lim_update2
      !!-------------------------------------------------------------------
      !!               ***  ROUTINE lim_update2  ***
      !!               
      !! ** Purpose :  Computes update of sea-ice global variables at 
      !!               the end of the time step.
      !!               Address pathological cases
      !!               This place is very important
      !!                
      !! ** Method  :  
      !!    Ice speed from ice dynamics
      !!    Ice thickness, Snow thickness, Temperatures, Lead fraction
      !!      from advection and ice thermodynamics 
      !!
      !! ** Action  : - 
      !!---------------------------------------------------------------------
      INTEGER ::   ji, jj, jk, jl, jm    ! dummy loop indices
      INTEGER ::   jbnd1, jbnd2
      INTEGER ::   i_ice_switch
      INTEGER ::   ind_im, layer      ! indices for internal melt
      REAL(wp) ::   zweight, zesum, zhimax, z_da_i
      REAL(wp) ::   zinda, zindb, zindsn, zindic
      REAL(wp) ::   zindg, zh, zdvres, zviold2
      REAL(wp) ::   zbigvalue, zvsold2, z_da_ex
      REAL(wp) ::   z_prescr_hi, zat_i_old, ztmelts, ze_s

      INTEGER , POINTER, DIMENSION(:,:,:) ::  internal_melt
      REAL(wp), POINTER, DIMENSION(:) ::   zthick0, zqm0      ! thickness of the layers and heat contents for
      REAL(wp) :: zchk_v_i, zchk_smv, zchk_fs, zchk_fw, zchk_v_i_b, zchk_smv_b, zchk_fs_b, zchk_fw_b ! Check conservation (C Rousset)
      REAL(wp) :: zchk_vmin, zchk_amin, zchk_amax ! Check errors (C Rousset)
      ! mass and salt flux (clem)
      REAL(wp), POINTER, DIMENSION(:,:,:) ::   zviold, zvsold, zsmvold   ! old ice volume...
      !!-------------------------------------------------------------------
      IF( nn_timing == 1 )  CALL timing_start('limupdate2')

      CALL wrk_alloc( jpi,jpj,jpl, internal_melt )   ! integer
      CALL wrk_alloc( jkmax, zthick0, zqm0 )

      CALL wrk_alloc( jpi,jpj,jpl,zviold, zvsold, zsmvold )   ! clem

      !----------------------------------------------------------------------------------------
      ! 1. Computation of trend terms      
      !----------------------------------------------------------------------------------------
      !- Trend terms
      d_a_i_thd(:,:,:)   = a_i(:,:,:)   - old_a_i(:,:,:) 
      d_v_s_thd(:,:,:)   = v_s(:,:,:)   - old_v_s(:,:,:)
      d_v_i_thd(:,:,:)   = v_i(:,:,:)   - old_v_i(:,:,:)  
      d_e_s_thd(:,:,:,:) = e_s(:,:,:,:) - old_e_s(:,:,:,:) 
      d_e_i_thd(:,:,:,:) = e_i(:,:,:,:) - old_e_i(:,:,:,:)
      !?? d_oa_i_thd(:,:,:)  = oa_i (:,:,:) - old_oa_i (:,:,:)
      d_smv_i_thd(:,:,:) = 0._wp
      IF( num_sal == 2 )   d_smv_i_thd(:,:,:) = smv_i(:,:,:) - old_smv_i(:,:,:)
      ! diag only (clem)
      dv_dt_thd(:,:,:) = d_v_i_thd(:,:,:) * r1_rdtice * rday

      ! mass and salt flux init (clem)
      zviold(:,:,:) = v_i(:,:,:)
      zvsold(:,:,:) = v_s(:,:,:)
      zsmvold(:,:,:) = smv_i(:,:,:)

      ! -------------------------------
      !- check conservation (C Rousset)
      IF (ln_limdiahsb) THEN
         zchk_v_i_b = glob_sum( SUM(   v_i(:,:,:), dim=3 ) * area(:,:) * tms(:,:) )
         zchk_smv_b = glob_sum( SUM( smv_i(:,:,:), dim=3 ) * area(:,:) * tms(:,:) )
         zchk_fw_b  = glob_sum( rdm_ice(:,:) * area(:,:) * tms(:,:) )
         zchk_fs_b  = glob_sum( ( sfx_bri(:,:) + sfx_thd(:,:) + sfx_res(:,:) + sfx_mec(:,:) ) * area(:,:) * tms(:,:) )
      ENDIF
      !- check conservation (C Rousset)
      ! -------------------------------

      CALL lim_var_glo2eqv

      !--------------------------------------
      ! 2. Review of all pathological cases
      !--------------------------------------

! clem: useless now
      !-------------------------------------------
      ! 2.1) Advection of ice in an ice-free cell
      !-------------------------------------------
      ! should be removed since it is treated after dynamics now
!      zhimax = 5._wp
!      ! first category
!      DO jj = 1, jpj
!         DO ji = 1, jpi
!            !--- the thickness of such an ice is often out of bounds
!            !--- thus we recompute a new area while conserving ice volume
!            zat_i_old = SUM( old_a_i(ji,jj,:) )
!            zindb     = MAX( 0._wp, SIGN( 1._wp, ABS( d_a_i_thd(ji,jj,1) ) - epsi10 ) ) 
!            IF ( ( ABS( d_v_i_thd(ji,jj,1) ) / MAX( ABS( d_a_i_thd(ji,jj,1) ),epsi10 ) * zindb .GT. zhimax ) &
!               &  .AND. ( ( v_i(ji,jj,1) / MAX( a_i(ji,jj,1), epsi10 ) * zindb ) .GT. zhimax ) &
!               &  .AND. ( zat_i_old .LT. 1.e-6 ) )  THEN ! new line
!               ht_i(ji,jj,1) = hi_max(1) * 0.5_wp
!               a_i (ji,jj,1) = v_i(ji,jj,1) / ht_i(ji,jj,1)
!            ENDIF
!         END DO
!      END DO

!      zhimax = 20._wp
!      ! other categories
!      DO jl = 2, jpl
!         jm = ice_types(jl)
!         DO jj = 1, jpj
!            DO ji = 1, jpi
!               zindb =  MAX( rzero, SIGN( rone, ABS( d_a_i_thd(ji,jj,jl)) - epsi10 ) ) 
!              ! this correction is very tricky... sometimes, advection gets wrong i don't know why
!               ! it makes problems when the advected volume and concentration do not seem to be 
!               ! related with each other
!               ! the new thickness is sometimes very big!
!               ! and sometimes d_a_i_trp and d_v_i_trp have different sign
!               ! which of course is plausible
!               ! but fuck! it fucks everything up :)
!               IF ( ( ABS( d_v_i_thd(ji,jj,jl) ) / MAX( ABS( d_a_i_thd(ji,jj,jl) ), epsi10 ) * zindb .GT. zhimax ) &
!                  &  .AND. ( v_i(ji,jj,jl) / MAX( a_i(ji,jj,jl), epsi10 ) * zindb ) .GT. zhimax ) THEN
!                  ht_i(ji,jj,jl) = ( hi_max_typ(jl-ice_cat_bounds(jm,1),jm) + hi_max_typ(jl-ice_cat_bounds(jm,1)+1,jm) ) * 0.5_wp
!                  a_i (ji,jj,jl) = v_i(ji,jj,jl) / ht_i(ji,jj,jl)
!               ENDIF
!            END DO ! ji
!         END DO !jj
!      END DO !jl
     
      at_i(:,:) = 0._wp
      DO jl = 1, jpl
         at_i(:,:) = a_i(:,:,jl) + at_i(:,:)
      END DO

      !----------------------------------------------------
      ! 2.2) Rebin categories with thickness out of bounds
      !----------------------------------------------------
      DO jm = 1, jpm
         jbnd1 = ice_cat_bounds(jm,1)
         jbnd2 = ice_cat_bounds(jm,2)
         IF (ice_ncat_types(jm) .GT. 1 )   CALL lim_itd_th_reb(jbnd1, jbnd2, jm)
      END DO

      !---------------------------------
      ! 2.3) Melt of an internal layer
      !---------------------------------
      internal_melt(:,:,:) = 0

      DO jl = 1, jpl
         DO jk = 1, nlay_i
            DO jj = 1, jpj 
               DO ji = 1, jpi
                  ztmelts = - tmut * s_i(ji,jj,jk,jl) + rtt
                  IF ( ( ( e_i(ji,jj,jk,jl) .LE. 0.0 ) .OR. ( t_i(ji,jj,jk,jl) .GE. ztmelts ) ) &
                    & .AND. ( v_i(ji,jj,jl) .GT. 0.0 ) .AND. ( a_i(ji,jj,jl) .GT. 0.0 ) ) THEN
                     internal_melt(ji,jj,jl) = 1
                  ENDIF
               END DO ! ji
            END DO ! jj
         END DO !jk
      END DO !jl

      DO jl = 1, jpl
         DO jj = 1, jpj 
            DO ji = 1, jpi
               IF( internal_melt(ji,jj,jl) == 1 ) THEN
                  ! initial ice thickness
                  !-----------------------
                  ht_i(ji,jj,jl) = v_i(ji,jj,jl) / a_i(ji,jj,jl)

                  ! reduce ice thickness
                  !-----------------------
                  ind_im = 0
                  zesum = 0.0
                  DO jk = 1, nlay_i
                     ztmelts = - tmut * s_i(ji,jj,jk,jl) + rtt
                     IF ( ( e_i(ji,jj,jk,jl) .LE. 0.0 ) .OR. ( t_i(ji,jj,jk,jl) .GE. ztmelts ) ) ind_im = ind_im + 1
                     zesum = zesum + e_i(ji,jj,jk,jl)
                  END DO
                  ht_i(ji,jj,jl) = ht_i(ji,jj,jl) - REAL(ind_im)*ht_i(ji,jj,jl) / REAL(nlay_i)
                  v_i(ji,jj,jl)  = ht_i(ji,jj,jl) * a_i(ji,jj,jl)

                  !CLEM
                  zdvres = REAL(ind_im)*ht_i(ji,jj,jl) / REAL(nlay_i) * a_i(ji,jj,jl)
                  !rdm_ice(ji,jj) = rdm_ice(ji,jj) - zdvres * rhoic
                  !sfx_res(ji,jj)  = sfx_res(ji,jj) + sm_i(ji,jj,jl) * ( rhoic * zdvres / rdt_ice )

                  ! redistribute heat
                  !-----------------------
                  ! old thicknesses and enthalpies
                  ind_im = 0
                  DO jk = 1, nlay_i
                     ztmelts = - tmut * s_i(ji,jj,jk,jl) + rtt
                     IF ( ( e_i(ji,jj,jk,jl) .GT. 0.0 ) .AND.  & 
                        ( t_i(ji,jj,jk,jl) .LT. ztmelts ) ) THEN
                        ind_im = ind_im + 1
                        zthick0(ind_im) = ht_i(ji,jj,jl) * REAL(ind_im / nlay_i)
                        zqm0   (ind_im) = MAX( e_i(ji,jj,jk,jl) , 0.0 )
                     ENDIF
                  END DO

                  ! Redistributing energy on the new grid
                  IF ( ind_im .GT. 0 ) THEN

                     DO jk = 1, nlay_i
                        e_i(ji,jj,jk,jl) = 0.0
                        DO layer = 1, ind_im
                           zweight         = MAX (  &
                              MIN( ht_i(ji,jj,jl) * REAL(layer/ind_im) , ht_i(ji,jj,jl) * REAL(jk / nlay_i) ) -       &
                              MAX( ht_i(ji,jj,jl) * REAL((layer-1)/ind_im) , ht_i(ji,jj,jl) * REAL((jk-1) / nlay_i) ) , 0.0 ) &
                              /  ( ht_i(ji,jj,jl) / REAL(ind_im) )

                           e_i(ji,jj,jk,jl) =  e_i(ji,jj,jk,jl) + zweight*zqm0(layer)
                        END DO !layer
                     END DO ! jk

                     zesum = 0.0
                     DO jk = 1, nlay_i
                        zesum = zesum + e_i(ji,jj,jk,jl)
                     END DO

                  ELSE ! ind_im .EQ. 0, total melt
                     e_i(ji,jj,jk,jl) = 0.0
                  ENDIF

               ENDIF ! internal_melt

            END DO ! ji
         END DO !jj
      END DO !jl

      internal_melt(:,:,:) = 0


      ! Melt of snow
      !--------------
      DO jl = 1, jpl
         DO jj = 1, jpj 
            DO ji = 1, jpi
               ! snow energy of melting
               zinda   =  MAX( 0._wp, SIGN( 1._wp, v_s(ji,jj,jl) - epsi10 ) )
               ze_s = zinda * e_s(ji,jj,1,jl) * unit_fac / area(ji,jj) / MAX( v_s(ji,jj,jl), epsi10 )  ! snow energy of melting

               ! If snow energy of melting smaller then Lf
               ! Then all snow melts and meltwater, heat go to the ocean
               IF ( ze_s .LE. rhosn * lfus ) internal_melt(ji,jj,jl) = 1

            END DO
         END DO
      END DO

      DO jl = 1, jpl
         DO jj = 1, jpj 
            DO ji = 1, jpi
               IF ( internal_melt(ji,jj,jl) == 1 ) THEN
                  zdvres = v_s(ji,jj,jl)
                  ! release heat
                  fheat_res(ji,jj) = fheat_res(ji,jj) + ze_s * zdvres / rdt_ice
                  ! release mass
                  !rdm_snw(ji,jj) =  rdm_snw(ji,jj) - zdvres * rhosn
                  !
                  v_s(ji,jj,jl)   = 0.0
                  e_s(ji,jj,1,jl) = 0.0
                 ENDIF
            END DO
         END DO
      END DO

      zbigvalue      = 1.0e+20
      DO jl = 1, jpl
         DO jj = 1, jpj 
            DO ji = 1, jpi

               !switches
               zindb         = MAX( rzero, SIGN( rone, a_i(ji,jj,jl) - epsi10 ) ) 
               !switch = 1 if a_i > 1e-06 and 0 if not
               zindsn        = MAX( rzero, SIGN( rone, v_s(ji,jj,jl) - epsi10 ) ) !=1 if hs > 1e-10 and 0 if not
               zindic        = MAX( rzero, SIGN( rone, v_i(ji,jj,jl) - epsi10 ) ) !=1 if hi > 1e-10 and 0 if not
               ! bug fix 25 avril 2007
               zindb         = zindb*zindic

               !--- 2.3 Correction to ice age 
               !------------------------------
               !                IF ((o_i(ji,jj,jl)-1.0)*rday.gt.(rdt_ice*float(numit))) THEN
               !                   o_i(ji,jj,jl) = rdt_ice*FLOAT(numit)/rday
               !                ENDIF
               IF ((oa_i(ji,jj,jl)-1.0)*rday.gt.(rdt_ice*numit*a_i(ji,jj,jl))) THEN
                  oa_i(ji,jj,jl) = rdt_ice*numit/rday*a_i(ji,jj,jl)
               ENDIF
               oa_i(ji,jj,jl) = zindb*zindic*oa_i(ji,jj,jl)

               !--- 2.4 Correction to snow thickness
               !-------------------------------------
               zdvres = (zindsn * zindb - 1._wp) * v_s(ji,jj,jl)
               v_s(ji,jj,jl) = v_s(ji,jj,jl) + zdvres

               !rdm_snw(ji,jj) = rdm_snw(ji,jj) + zdvres * rhosn
 
               !--- 2.5 Correction to ice thickness
               !-------------------------------------
               zdvres = (zindb - 1._wp) * v_i(ji,jj,jl)
               v_i(ji,jj,jl) = v_i(ji,jj,jl) + zdvres

               !rdm_ice(ji,jj) = rdm_ice(ji,jj) + zdvres * rhoic
               !sfx_res(ji,jj)  = sfx_res(ji,jj) - sm_i(ji,jj,jl) * ( rhoic * zdvres / rdt_ice )

               !--- 2.6 Snow is transformed into ice if the original ice cover disappears
               !----------------------------------------------------------------------------
               zindg          = tms(ji,jj) *  MAX( 0._wp, SIGN( 1._wp, -v_i(ji,jj,jl) ) )
               zdvres         = zindg * rhosn * v_s(ji,jj,jl) / rau0
               v_i(ji,jj,jl)  = v_i(ji,jj,jl)  + zdvres

               zdvres         = zindsn*zindb * ( - zindg * v_s(ji,jj,jl) + zindg * v_i(ji,jj,jl) * ( rau0 - rhoic ) / rhosn )
               v_s(ji,jj,jl)  = v_s(ji,jj,jl)  + zdvres

               !--- 2.7 Correction to ice concentrations 
               !--------------------------------------------
               a_i(ji,jj,jl) = zindb * a_i(ji,jj,jl)

               !-------------------------
               ! 2.8) Snow heat content
               !-------------------------
               e_s(ji,jj,1,jl) = zindsn * ( MIN ( MAX ( 0.0, e_s(ji,jj,1,jl) ), zbigvalue ) )

            END DO ! ji
         END DO ! jj
      END DO ! jl

      !------------------------
      ! 2.9) Ice heat content 
      !------------------------

      DO jl = 1, jpl
         DO jk = 1, nlay_i
            DO jj = 1, jpj 
               DO ji = 1, jpi
                  zindic        = MAX( rzero, SIGN( rone, v_i(ji,jj,jl) - epsi10 ) ) 
                  e_i(ji,jj,jk,jl)= zindic * ( MIN ( MAX ( 0.0, e_i(ji,jj,jk,jl) ), zbigvalue ) )
               END DO ! ji
            END DO ! jj
         END DO !jk
      END DO !jl


      DO jm = 1, jpm
         DO jj = 1, jpj 
            DO ji = 1, jpi
               jl = ice_cat_bounds(jm,1)
               !--- 2.12 Constrain the thickness of the smallest category above 5 cm
               !----------------------------------------------------------------------
               zindb         = MAX( rzero, SIGN( rone, a_i(ji,jj,jl) - epsi10 ) ) 
               ht_i(ji,jj,jl) = zindb*v_i(ji,jj,jl)/MAX(a_i(ji,jj,jl), epsi10)
               zh            = MAX( rone , zindb * hiclim  / MAX( ht_i(ji,jj,jl) , epsi10 ) )
               ht_s(ji,jj,jl) = ht_s(ji,jj,jl)* zh
               ht_i(ji,jj,jl) = ht_i(ji,jj,jl)* zh
               a_i (ji,jj,jl) = a_i(ji,jj,jl) / zh
               !CLEM
               v_i (ji,jj,jl) = a_i(ji,jj,jl) * ht_i(ji,jj,jl)
               v_s (ji,jj,jl) = a_i(ji,jj,jl) * ht_s(ji,jj,jl)
            END DO !ji
         END DO !jj
      END DO !jm

      at_i(:,:) = 0.0
      DO jl = 1, jpl
         at_i(:,:) = a_i(:,:,jl) + at_i(:,:)
      END DO
      
      !--- 2.13 ice concentration should not exceed amax 
      !         (it should not be the case) 
      !-----------------------------------------------------
      DO jj = 1, jpj
         DO ji = 1, jpi
            z_da_ex =  MAX( at_i(ji,jj) - amax , 0.0 )        
            zindb   =  MAX( rzero, SIGN( rone, at_i(ji,jj) - epsi10 ) ) 
            DO jl  = 1, jpl
               z_da_i = a_i(ji,jj,jl) * z_da_ex / MAX( at_i(ji,jj), epsi10 ) * zindb
               a_i(ji,jj,jl) = MAX( 0._wp, a_i(ji,jj,jl) - z_da_i )
               !
               zinda   =  MAX( rzero, SIGN( rone, a_i(ji,jj,jl) - epsi10 ) ) 
               ht_i(ji,jj,jl) = v_i(ji,jj,jl) / MAX( a_i(ji,jj,jl), epsi10 ) * zinda
               !v_i(ji,jj,jl) = ht_i(ji,jj,jl) * a_i(ji,jj,jl) ! makes ice shrinken but should not be used
            END DO
         END DO
      END DO
      at_i(:,:) = 0.0
      DO jl = 1, jpl
         at_i(:,:) = a_i(:,:,jl) + at_i(:,:)
      END DO

      ! Final thickness distribution rebinning
      ! --------------------------------------
      DO jm = 1, jpm
         jbnd1 = ice_cat_bounds(jm,1)
         jbnd2 = ice_cat_bounds(jm,2)
         IF (ice_ncat_types(jm) .GT. 1 ) CALL lim_itd_th_reb(jbnd1, jbnd2, jm)
         IF (ice_ncat_types(jm) .EQ. 1 ) THEN
         ENDIF
      END DO

      !---------------------
      ! 2.11) Ice salinity
      !---------------------
      ! clem correct bug on smv_i
      smv_i(:,:,:) = sm_i(:,:,:) * v_i(:,:,:)

      IF (  num_sal == 2  ) THEN ! general case
         DO jl = 1, jpl
            !DO jk = 1, nlay_i
               DO jj = 1, jpj 
                  DO ji = 1, jpi
                     ! salinity stays in bounds
                     !clem smv_i(ji,jj,jl)  =  MAX(MIN((rhoic-rhosn)/rhoic*sss_m(ji,jj),smv_i(ji,jj,jl)),0.1 * v_i(ji,jj,jl) )
                     smv_i(ji,jj,jl) = MAX( MIN( s_i_max * v_i(ji,jj,jl), smv_i(ji,jj,jl) ), s_i_min * v_i(ji,jj,jl) )
                     i_ice_switch    = 1._wp - MAX( 0._wp, SIGN( 1._wp, -v_i(ji,jj,jl) ) )
                     smv_i(ji,jj,jl) = i_ice_switch * smv_i(ji,jj,jl) !+ s_i_min * ( 1._wp - i_ice_switch ) * v_i(ji,jj,jl)
                  END DO ! ji
               END DO ! jj
            !END DO !jk
         END DO !jl
      ENDIF

      ! -------------------
      at_i(:,:) = a_i(:,:,1)
      DO jl = 2, jpl
         at_i(:,:) = a_i(:,:,jl) + at_i(:,:)
      END DO

      !------------------------------------------------------------------------------
      ! 2) Corrections to avoid wrong values                                        |
      !------------------------------------------------------------------------------
      ! Ice drift
      !------------
      DO jj = 2, jpjm1
         DO ji = 2, jpim1
            IF ( at_i(ji,jj) .EQ. 0.0 ) THEN ! what to do if there is no ice
               IF ( at_i(ji+1,jj) .EQ. 0.0 ) u_ice(ji,jj)   = 0.0 ! right side
               IF ( at_i(ji-1,jj) .EQ. 0.0 ) u_ice(ji-1,jj) = 0.0 ! left side
               IF ( at_i(ji,jj+1) .EQ. 0.0 ) v_ice(ji,jj)   = 0.0 ! upper side
               IF ( at_i(ji,jj-1) .EQ. 0.0 ) v_ice(ji,jj-1) = 0.0 ! bottom side
            ENDIF
         END DO
      END DO
      !lateral boundary conditions
      CALL lbc_lnk( u_ice(:,:), 'U', -1. )
      CALL lbc_lnk( v_ice(:,:), 'V', -1. )
      !mask velocities
      u_ice(:,:) = u_ice(:,:) * tmu(:,:)
      v_ice(:,:) = v_ice(:,:) * tmv(:,:)
 
      !--------------------------------
      ! Update mass/salt fluxes (clem)
      !--------------------------------
      DO jl = 1, jpl
         DO jj = 1, jpj 
            DO ji = 1, jpi
               diag_res_pr(ji,jj) = diag_res_pr(ji,jj) + ( v_i(ji,jj,jl) - zviold(ji,jj,jl) ) / rdt_ice 
               rdm_ice(ji,jj) = rdm_ice(ji,jj) + ( v_i(ji,jj,jl) - zviold(ji,jj,jl) ) * rhoic 
               rdm_snw(ji,jj) = rdm_snw(ji,jj) + ( v_s(ji,jj,jl) - zvsold(ji,jj,jl) ) * rhosn 
               sfx_res(ji,jj) = sfx_res(ji,jj) - ( smv_i(ji,jj,jl) - zsmvold(ji,jj,jl) ) * rhoic / rdt_ice 
            END DO
         END DO
      END DO

      ! -------------------------------
      !- check conservation (C Rousset)
      IF (ln_limdiahsb) THEN

         zchk_fs  = glob_sum( ( sfx_bri(:,:) + sfx_thd(:,:) + sfx_res(:,:) + sfx_mec(:,:) ) * area(:,:) * tms(:,:) ) - zchk_fs_b
         zchk_fw  = glob_sum( rdm_ice(:,:) * area(:,:) * tms(:,:) ) - zchk_fw_b
 
         zchk_v_i = ( glob_sum( SUM(   v_i(:,:,:), dim=3 ) * area(:,:) * tms(:,:) ) - zchk_v_i_b - ( zchk_fw / rhoic ) ) * r1_rdtice
         zchk_smv = ( glob_sum( SUM( smv_i(:,:,:), dim=3 ) * area(:,:) * tms(:,:) ) - zchk_smv_b ) * r1_rdtice + ( zchk_fs / rhoic )

         zchk_vmin = glob_min(v_i)
         zchk_amax = glob_max(SUM(a_i,dim=3))
         zchk_amin = glob_min(a_i)

         IF(lwp) THEN
            IF ( ABS( zchk_v_i   ) >  1.e-5 ) WRITE(numout,*) 'violation volume [m3/day]     (limupdate2) = ',(zchk_v_i * rday)
            IF ( ABS( zchk_smv   ) >  1.e-4 ) WRITE(numout,*) 'violation saline [psu*m3/day] (limupdate2) = ',(zchk_smv * rday)
            IF ( zchk_vmin <  0.            ) WRITE(numout,*) 'violation v_i<0  [mm]         (limupdate2) = ',(zchk_vmin * 1.e-3)
            IF ( zchk_amax >  amax+epsi10   ) WRITE(numout,*) 'violation a_i>amax            (limupdate2) = ',zchk_amax
            IF ( zchk_amin <  0.            ) WRITE(numout,*) 'violation a_i<0               (limupdate2) = ',zchk_amin
         ENDIF
      ENDIF
      !- check conservation (C Rousset)
      ! -------------------------------

      IF(ln_ctl) THEN   ! Control print
         CALL prt_ctl_info(' ')
         CALL prt_ctl_info(' - Cell values : ')
         CALL prt_ctl_info('   ~~~~~~~~~~~~~ ')
         CALL prt_ctl(tab2d_1=area       , clinfo1=' lim_update2  : cell area   :')
         CALL prt_ctl(tab2d_1=at_i       , clinfo1=' lim_update2  : at_i        :')
         CALL prt_ctl(tab2d_1=vt_i       , clinfo1=' lim_update2  : vt_i        :')
         CALL prt_ctl(tab2d_1=vt_s       , clinfo1=' lim_update2  : vt_s        :')
         CALL prt_ctl(tab2d_1=strength   , clinfo1=' lim_update2  : strength    :')
         CALL prt_ctl(tab2d_1=u_ice      , clinfo1=' lim_update2  : u_ice       :', tab2d_2=v_ice      , clinfo2=' v_ice       :')
         CALL prt_ctl(tab2d_1=old_u_ice  , clinfo1=' lim_update2  : old_u_ice   :', tab2d_2=old_v_ice  , clinfo2=' old_v_ice   :')

         DO jl = 1, jpl
            CALL prt_ctl_info(' ')
            CALL prt_ctl_info(' - Category : ', ivar1=jl)
            CALL prt_ctl_info('   ~~~~~~~~~~')
            CALL prt_ctl(tab2d_1=ht_i       (:,:,jl)        , clinfo1= ' lim_update2  : ht_i        : ')
            CALL prt_ctl(tab2d_1=ht_s       (:,:,jl)        , clinfo1= ' lim_update2  : ht_s        : ')
            CALL prt_ctl(tab2d_1=t_su       (:,:,jl)        , clinfo1= ' lim_update2  : t_su        : ')
            CALL prt_ctl(tab2d_1=t_s        (:,:,1,jl)      , clinfo1= ' lim_update2  : t_snow      : ')
            CALL prt_ctl(tab2d_1=sm_i       (:,:,jl)        , clinfo1= ' lim_update2  : sm_i        : ')
            CALL prt_ctl(tab2d_1=o_i        (:,:,jl)        , clinfo1= ' lim_update2  : o_i         : ')
            CALL prt_ctl(tab2d_1=a_i        (:,:,jl)        , clinfo1= ' lim_update2  : a_i         : ')
            CALL prt_ctl(tab2d_1=old_a_i    (:,:,jl)        , clinfo1= ' lim_update2  : old_a_i     : ')
            CALL prt_ctl(tab2d_1=d_a_i_thd  (:,:,jl)        , clinfo1= ' lim_update2  : d_a_i_thd   : ')
            CALL prt_ctl(tab2d_1=v_i        (:,:,jl)        , clinfo1= ' lim_update2  : v_i         : ')
            CALL prt_ctl(tab2d_1=old_v_i    (:,:,jl)        , clinfo1= ' lim_update2  : old_v_i     : ')
            CALL prt_ctl(tab2d_1=d_v_i_thd  (:,:,jl)        , clinfo1= ' lim_update2  : d_v_i_thd   : ')
            CALL prt_ctl(tab2d_1=v_s        (:,:,jl)        , clinfo1= ' lim_update2  : v_s         : ')
            CALL prt_ctl(tab2d_1=old_v_s    (:,:,jl)        , clinfo1= ' lim_update2  : old_v_s     : ')
            CALL prt_ctl(tab2d_1=d_v_s_thd  (:,:,jl)        , clinfo1= ' lim_update2  : d_v_s_thd   : ')
            CALL prt_ctl(tab2d_1=e_i        (:,:,1,jl)/1.0e9, clinfo1= ' lim_update2  : e_i1        : ')
            CALL prt_ctl(tab2d_1=old_e_i    (:,:,1,jl)/1.0e9, clinfo1= ' lim_update2  : old_e_i1    : ')
            CALL prt_ctl(tab2d_1=d_e_i_thd  (:,:,1,jl)/1.0e9, clinfo1= ' lim_update2  : de_i1_thd   : ')
            CALL prt_ctl(tab2d_1=e_i        (:,:,2,jl)/1.0e9, clinfo1= ' lim_update2  : e_i2        : ')
            CALL prt_ctl(tab2d_1=old_e_i    (:,:,2,jl)/1.0e9, clinfo1= ' lim_update2  : old_e_i2    : ')
            CALL prt_ctl(tab2d_1=d_e_i_thd  (:,:,2,jl)/1.0e9, clinfo1= ' lim_update2  : de_i2_thd   : ')
            CALL prt_ctl(tab2d_1=e_s        (:,:,1,jl)      , clinfo1= ' lim_update2  : e_snow      : ')
            CALL prt_ctl(tab2d_1=old_e_s    (:,:,1,jl)      , clinfo1= ' lim_update2  : old_e_snow  : ')
            CALL prt_ctl(tab2d_1=d_e_s_thd  (:,:,1,jl)/1.0e9, clinfo1= ' lim_update2  : d_e_s_thd   : ')
            CALL prt_ctl(tab2d_1=smv_i      (:,:,jl)        , clinfo1= ' lim_update2  : smv_i       : ')
            CALL prt_ctl(tab2d_1=old_smv_i  (:,:,jl)        , clinfo1= ' lim_update2  : old_smv_i   : ')
            CALL prt_ctl(tab2d_1=d_smv_i_thd(:,:,jl)        , clinfo1= ' lim_update2  : d_smv_i_thd : ')
            CALL prt_ctl(tab2d_1=oa_i       (:,:,jl)        , clinfo1= ' lim_update2  : oa_i        : ')
            CALL prt_ctl(tab2d_1=old_oa_i   (:,:,jl)        , clinfo1= ' lim_update2  : old_oa_i    : ')
            CALL prt_ctl(tab2d_1=d_oa_i_thd (:,:,jl)        , clinfo1= ' lim_update2  : d_oa_i_thd  : ')

            DO jk = 1, nlay_i
               CALL prt_ctl_info(' - Layer : ', ivar1=jk)
               CALL prt_ctl(tab2d_1=t_i(:,:,jk,jl) , clinfo1= ' lim_update2  : t_i       : ')
            END DO
         END DO

         CALL prt_ctl_info(' ')
         CALL prt_ctl_info(' - Heat / FW fluxes : ')
         CALL prt_ctl_info('   ~~~~~~~~~~~~~~~~~~ ')
         CALL prt_ctl(tab2d_1=fmmec  , clinfo1= ' lim_update2 : fmmec : ', tab2d_2=fhmec     , clinfo2= ' fhmec     : ')
         CALL prt_ctl(tab2d_1=sst_m  , clinfo1= ' lim_update2 : sst   : ', tab2d_2=sss_m     , clinfo2= ' sss       : ')
         CALL prt_ctl(tab2d_1=fhbri  , clinfo1= ' lim_update2 : fhbri : ')

         CALL prt_ctl_info(' ')
         CALL prt_ctl_info(' - Stresses : ')
         CALL prt_ctl_info('   ~~~~~~~~~~ ')
         CALL prt_ctl(tab2d_1=utau       , clinfo1= ' lim_update2 : utau      : ', tab2d_2=vtau       , clinfo2= ' vtau      : ')
         CALL prt_ctl(tab2d_1=utau_ice   , clinfo1= ' lim_update2 : utau_ice  : ', tab2d_2=vtau_ice   , clinfo2= ' vtau_ice  : ')
         CALL prt_ctl(tab2d_1=u_oce      , clinfo1= ' lim_update2 : u_oce     : ', tab2d_2=v_oce      , clinfo2= ' v_oce     : ')
      ENDIF
   
      CALL wrk_dealloc( jpi,jpj,jpl, internal_melt )   ! integer
      CALL wrk_dealloc( jkmax, zthick0, zqm0 )

      CALL wrk_dealloc( jpi,jpj,jpl,zviold, zvsold, zsmvold )   ! clem

      IF( nn_timing == 1 )  CALL timing_stop('limupdate2')
   END SUBROUTINE lim_update2
#else
   !!----------------------------------------------------------------------
   !!   Default option         Empty Module               No sea-ice model
   !!----------------------------------------------------------------------
CONTAINS
   SUBROUTINE lim_update2     ! Empty routine
   END SUBROUTINE lim_update2

#endif

END MODULE limupdate2
